//Written by Jiangjian Xiao

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "img.h"
#include "math.h"


int round(float a) 
{
	int b=(int)floor(a);
	if(a-b>0.5)
		return b+1;
	else
		return b;
}

int round(double a) 
{
	int b=(int)floor(a);
	if(a-b>0.5)
		return b+1;
	else
		return b;
}


PgmImage::PgmImage(int _row,int _col) 
{
	int r,c;
	row=_row;
	col=_col;
	img = new uchar*[row];		//assign memory to source and target arrays
	
	for (r=0; r<row; r++)		//initialize image to 0
	{
		img[r]=new uchar[col];
		for (c=0; c<col; c++)
		{
			img[r][c] = 0;
		}
	}
}

PgmImage::PgmImage(int _row,int _col, uchar color) 
{
	int r,c;
	row=_row;
	col=_col;
	img = new uchar*[row];		//assign memory to source and target arrays
	
	for (r=0; r<row; r++)		//initialize image to 0
	{
		img[r]=new uchar[col];
		for (c=0; c<col; c++)
		{
			img[r][c] = color;
		}
	}
}

PgmImage::PgmImage(PgmImage *_img)
{
	int r,c;
	row=_img->row;
	col=_img->col;
	img = new uchar*[row];		//assign memory to source and target arrays
	
	for (r=0; r<row; r++)		//initialize image to 0
	{
		img[r]=new uchar[col];
		for (c=0; c<col; c++)
		{
			img[r][c] = _img->img[r][c];
		}
	}
}

PgmImage::PgmImage()
{
	row=-1;
	col=-1;
	img=NULL;
}

PgmImage::PgmImage(char *filename)
{
	img=NULL;
	ReadPGM(filename);
}

PgmImage::~PgmImage()
{
	int i;

	if(row!=-1)
	{
		for(i=0; i<row; i++)
		{
			delete img[i];
		}
		delete img;
	}
}

void PgmImage::Copy(PgmImage* _img)
{
	int r,c;

	for (r=0; r<row; r++)		//initialize image to 0
	{
		for (c=0; c<col; c++)
		{
			img[r][c] = _img->img[r][c];
		}
	}
}

void PgmImage::Clear()
{
	int r,c;

	for (r=0; r<row; r++)		//initialize image to 0
	{
		for (c=0; c<col; c++)
		{
			img[r][c] = 0;
		}
	}
}

void PgmImage::Clear(uchar color)
{
	int r,c;

	for (r=0; r<row; r++)		//initialize image to 0
	{
		for (c=0; c<col; c++)
		{
			img[r][c] = color;
		}
	}
}

void PgmImage::ReadPGM(char *filename)
{
	FILE *fp;
	int r,c;
	char buff[256];
	int gray;
	int pgmtype;

	if ((fp = fopen(filename,"rb")) == NULL)
	{
		printf("File does not exist!\n");	//bad source file name
		exit(0);
	}
	else
	{
		if(fgetc(fp)!='P')			//check if the source file is in correct format
		{	
			printf("Not a PGM file!\n");
			fclose(fp);
			return;
		}
		else
		{	
			fgets(buff,256,fp);		//inputs the first line in the image (starting from 'P')
			if(buff[0]=='2')
			{
				pgmtype=2;	
			}
			else
			{
				pgmtype=5;
			}

			while(1)
			{
				fgets(buff,256,fp);	//comments are stored in char array comments
				if(buff[0]!='#')
				{
					sscanf(buff,"%d %d",&col,&row);
					break;
				}
			}
			
			fgets(buff,256,fp);		//read max grey level
			sscanf(buff,"%d",&colorLevel);
			
			if(img==NULL)
			{
				img = new uchar*[row];		//assign memory to source and target arrays

				for (r=0; r<row; r++)
				{
					img[r]=new uchar[col];
				}
			}
			
			if(pgmtype==5)					//binary file
			{
				for (r=0; r<row; r++)	//read image from the file and loads in the array
					fread(img[r],sizeof(uchar),col,fp);
			}
			else
			{
				for (r=0;r<row;r++)		//ascii file
				{
					for(c=0;c<col-1;c++)
					{
						fscanf(fp,"%d",&gray);
						img[r][c]=gray;
					}
					fscanf(fp,"%d\n",&gray);
					img[r][col-1]=gray;
				}
			}
			fclose(fp);					//close the file pointer
		}		
	}
}

void PgmImage::Edge(float ht,float lt,float sigma)
{
	int r,c,i,j;

	float **gaussMaskX;
	float **gaussMaskY;
	float Sx, Sy;				//variables to hold partial derivatives' values
	int diameter;
	int radius;
	int **dir;	
	float	**im1,**im2,**mag,**im4;


	im1 = new float*[row];		//assign memory to source and target arrays
	im2 = new float*[row];
	mag = new float*[row];
	im4 = new float*[row];
	dir = new int*[row];
	for (r=0; r<row; r++)
	{
		im1[r]=new float[col];
		im2[r]=new float[col];
		mag[r]=new float[col];
		im4[r]=new float[col];
		dir[r]=new int[col];
	}

	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{	
			im1[r][c]=img[r][c];
		}
	}

	float ss=sigma*sigma;
	radius=(int)round(sqrt(-log(0.005)*2*ss));
	diameter=2*radius+1;
	/*
	for(i=0;i<50;i++)
	{
		if(exp((-i*i)/(2*ss))<0.01)
		{
			diameter=2*(i-1)+1;
			break;
		}
	}
	radius=diameter/2;
	*/

	gaussMaskX = new float*[diameter];		//assign memory to source and target arrays
	gaussMaskY = new float*[diameter];
	for (i=0; i<diameter; i++)
	{
		gaussMaskX[i]=new float[diameter];
		gaussMaskY[i]=new float[diameter];
	}

	for(i=-radius;i<=radius;i++)
	{
		for(j=-radius;j<=radius;j++)
		{
			gaussMaskY[i+radius][j+radius]=(float)(-(-i/(ss))*exp(-(i*i+j*j)/(2.0*ss)));
			gaussMaskX[i+radius][j+radius]=(float)(-(j/(ss))*exp(-(i*i+j*j)/(2.0*ss)));
		}
	}

	
	float max=0;
	float tan1_8=(float)tan(PI/8);
	float tan3_8=(float)tan(PI*3/8);
	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{	
			Sx=0;						//initialize derivative values
			Sy=0;
			for (i=0; i<diameter; i++)		//convolve the mask
			{
				for (j=0; j<diameter; j++)
				{
					int ri=r+i-radius;
					int cj=c+j-radius;
					if(ri>=0 && ri<row && cj>=0 && cj<col)
					{
						Sx = Sx+ im1[ri][cj]*gaussMaskX[i][j];
						Sy = Sy+ im1[ri][cj]*gaussMaskY[i][j];
					}
				}
			}
			/*
			if(Sx==0)
			{
				if(Sy>0)
				{
					dir[r][c]=PI/2;
				}
				else if(Sy<0)
				{
					dir[r][c]=-PI/2;
				}
				else
					dir[r][c]=0;
			}
			else
			{
				dir[r][c]=atan(Sy/Sx);
			}*/
			//
			//                         The X marks the pixel in question, and each
			//     4      1      2     of the quadrants for the gradient vector
			//       O----0----0       fall into two cases, divided by the 45 
			//       |         |       degree line.  In one case the gradient
			//       |         |       vector is more horizontal, and in the other
			//     3 O    X    O 3     it is more vertical.  There are eight 
			//       |         |       divisions, but for the non-maximum supression  
			//       |         |       we are only worried about 4 of them since we 
			//       O----O----O       use symmetric points about the center pixel.
			//     2      1      4    

			if(Sx==0)
			{
				dir[r][c]=1;
			}
			else
			{
				float t=Sy/Sx;
				if(t>tan1_8 && t<=tan3_8 )
				{
					dir[r][c]=2;
				}
				else if(t>-tan1_8 && t<=tan1_8 )
				{
					dir[r][c]=3;
				}
				else if(t>-tan3_8 && t<=-tan1_8 )
				{
					dir[r][c]=4;
				}
				else 				
				{
					dir[r][c]=1;
				}
			}
			
			mag[r][c]=(float)sqrt(Sx*Sx+Sy*Sy);
			if(max<mag[r][c])
			{
				max=mag[r][c];
			}
		}
	}

	
	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{
			mag[r][c]/=max;
		}
	}

	float Pi1_8=(float)PI/8;
	float Pi3_8=(float)PI*3/8;
	for(r=1; r<row-1; r++)
	{
		for(c=1; c<col-1; c++)
		{	
			switch(dir[r][c])
			{
				case 1:
            	if(mag[r][c]>mag[r-1][c] && mag[r][c]>=mag[r+1][c])
            		im4[r][c]=mag[r][c];
            	else
            		im4[r][c]=0;
					break;
				case 2:
            	if(mag[r][c]>mag[r-1][c+1] && mag[r][c]>=mag[r+1][c-1])	
            		im4[r][c]=mag[r][c];
            	else
            		im4[r][c]=0;
					break;
				case 3:
            	if(mag[r][c]>mag[r][c-1] && mag[r][c]>=mag[r][c+1])
            		im4[r][c]=mag[r][c];
            	else
            		im4[r][c]=0;
 					break;
				case 4:
            	if(mag[r][c]>mag[r-1][c-1] && mag[r][c]>=mag[r+1][c+1])
            		im4[r][c]=mag[r][c];
            	else
            		im4[r][c]=0;
					break;
			}
		}
	}

	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{	
			im2[r][c]=255;
		}
	}

	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{	
			if(im4[r][c]>ht)
			{
				im2[r][c]=0;		//mark the pixel if derivative is greater than threshold
			}
			else
			{
				im2[r][c]=255;	
			}
		}
	}
	
	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{
			if(im2[r][c]==0)
			{
				Connect(r, c, lt, im2, im4);
			}
		}
	}

	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{
			if(im2[r][c]==0)
			{
				img[r][c]=255;
			}
			else 
			{
				img[r][c]=0;
			}
		}
	}

	for(r=0; r<row; r++)
	{
		delete im1[r];
		delete im2[r];
		delete mag[r];
		delete im4[r];
		delete dir[r];
	}
	delete im1;
	delete im2;
	delete mag;
	delete im4;
	delete dir;


	for(i=0; i<diameter; i++)
	{
		delete gaussMaskX[i];
		delete gaussMaskY[i];
	}
	delete gaussMaskX;
	delete gaussMaskY;
}

void PgmImage::VerticalEdge(float ht,float lt,float sigma)
{
	int r,c,i,j;

	float **gaussMaskX;
	float **gaussMaskY;
	float Sx, Sy;				//variables to hold partial derivatives' values
	int diameter;
	int radius;
	int **dir;	
	float	**im1,**im2,**mag,**im4;


	im1 = new float*[row];		//assign memory to source and target arrays
	im2 = new float*[row];
	mag = new float*[row];
	im4 = new float*[row];
	dir = new int*[row];
	for (r=0; r<row; r++)
	{
		im1[r]=new float[col];
		im2[r]=new float[col];
		mag[r]=new float[col];
		im4[r]=new float[col];
		dir[r]=new int[col];
	}

	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{	
			im1[r][c]=img[r][c];
		}
	}

	float ss=sigma*sigma;
	for(i=0;i<50;i++)
	{
		if(exp((-i*i)/(2*ss))<0.01)
		{
			diameter=2*(i-1)+1;
			break;
		}
	}
	radius=diameter/2;

	gaussMaskX = new float*[diameter];		//assign memory to source and target arrays
	gaussMaskY = new float*[diameter];
	for (i=0; i<diameter; i++)
	{
		gaussMaskX[i]=new float[diameter];
		gaussMaskY[i]=new float[diameter];
	}

	for(i=-radius;i<=radius;i++)
	{
		for(j=-radius;j<=radius;j++)
		{
			gaussMaskY[i+radius][j+radius]=(float)(-(-i/(ss))*exp(-(i*i+j*j)/(2.0*ss)));
			gaussMaskX[i+radius][j+radius]=(float)(-(j/(ss))*exp(-(i*i+j*j)/(2.0*ss)));
		}
	}

	
	float max=0;
	float tan1_8=(float)tan(PI/8);
	float tan3_8=(float)tan(PI*3/8);
	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{	
			Sx=0;						//initialize derivative values
			Sy=0;
			for (i=0; i<diameter; i++)		//convolve the mask
			{
				for (j=0; j<diameter; j++)
				{
					int ri=r+i-radius;
					int cj=c+j-radius;
					if(ri>=0 && ri<row && cj>=0 && cj<col)
					{
						Sx = Sx+ im1[ri][cj]*gaussMaskX[i][j];
						Sy = Sy+ im1[ri][cj]*gaussMaskY[i][j];
					}
				}
			}
			/*
			if(Sx==0)
			{
				if(Sy>0)
				{
					dir[r][c]=PI/2;
				}
				else if(Sy<0)
				{
					dir[r][c]=-PI/2;
				}
				else
					dir[r][c]=0;
			}
			else
			{
				dir[r][c]=atan(Sy/Sx);
			}*/
			//
			//                         The X marks the pixel in question, and each
			//     4      1      2     of the quadrants for the gradient vector
			//       O----0----0       fall into two cases, divided by the 45 
			//       |         |       degree line.  In one case the gradient
			//       |         |       vector is more horizontal, and in the other
			//     3 O    X    O 3     it is more vertical.  There are eight 
			//       |         |       divisions, but for the non-maximum supression  
			//       |         |       we are only worried about 4 of them since we 
			//       O----O----O       use symmetric points about the center pixel.
			//     2      1      4    

			if(Sx==0)
			{
				dir[r][c]=1;
			}
			else
			{
				float t=Sy/Sx;
				if(t>tan1_8 && t<=tan3_8 )
				{
					dir[r][c]=2;
				}
				else if(t>-tan1_8 && t<=tan1_8 )
				{
					dir[r][c]=3;
				}
				else if(t>-tan3_8 && t<=-tan1_8 )
				{
					dir[r][c]=4;
				}
				else 				
				{
					dir[r][c]=1;
				}
			}
			
         mag[r][c]=(float)sqrt(Sx*Sx+Sy*Sy);
			if(max<mag[r][c])
			{
				max=mag[r][c];
			}
		}
	}

	
	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{
			mag[r][c]/=max;
		}
	}

	float Pi1_8=(float)PI/8;
	float Pi3_8=(float)PI*3/8;
	for(r=1; r<row-1; r++)
	{
		for(c=1; c<col-1; c++)
		{	
			switch(dir[r][c])
			{
				case 1:
            	//if(mag[r][c]>mag[r-1][c] && mag[r][c]>=mag[r+1][c])
            	//	im4[r][c]=mag[r][c];
            	//else
            		im4[r][c]=0;
					break;
				case 2:
            	if(mag[r][c]>mag[r-1][c+1] && mag[r][c]>=mag[r+1][c-1])	
            		im4[r][c]=mag[r][c];
            	else
            		im4[r][c]=0;
					break;
				case 3:
            	if(mag[r][c]>mag[r][c-1] && mag[r][c]>=mag[r][c+1])
            		im4[r][c]=mag[r][c];
            	else
            		im4[r][c]=0;
 					break;
				case 4:
            	if(mag[r][c]>mag[r-1][c-1] && mag[r][c]>=mag[r+1][c+1])
            		im4[r][c]=mag[r][c];
            	else
            		im4[r][c]=0;
					break;
			}
		}
	}

	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{	
			im2[r][c]=255;
		}
	}

	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{	
			if(im4[r][c]>ht)
			{
				im2[r][c]=0;		//mark the pixel if derivative is greater than threshold
			}
			else
			{
				im2[r][c]=255;	
			}
		}
	}
	
	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{
			if(im2[r][c]==0)
			{
				Connect(r, c, lt, im2, im4);
			}
		}
	}

	for(r=0; r<row; r++)
	{
		for(c=0; c<col; c++)
		{
			if(im2[r][c]==0)
			{
				img[r][c]=255;
			}
			else 
			{
				img[r][c]=0;
			}
		}
	}

	for(r=0; r<row; r++)
	{
		delete im1[r];
		delete im2[r];
		delete mag[r];
		delete im4[r];
		delete dir[r];
	}
	delete im1;
	delete im2;
	delete mag;
	delete im4;
	delete dir;


	for(i=0; i<diameter; i++)
	{
		delete gaussMaskX[i];
		delete gaussMaskY[i];
	}
	delete gaussMaskX;
	delete gaussMaskY;
}


void PgmImage::Connect(int r, int c, float lt, float **im2, float **im4)
{
	if(r-1>=0&&r+1<row&&c-1>=0&&c+1<col)
	{
		if(im2[r-1][c]!=0&&im4[r-1][c]>lt)
		{
			im2[r-1][c]=0;
			Connect(r-1, c, lt, im2, im4);
		}
		
		if(im2[r-1][c-1]!=0&&im4[r-1][c-1]>lt)
		{
			im2[r-1][c-1]=0;
			Connect(r-1, c-1, lt, im2, im4);
		}
		
		if(im2[r-1][c+1]!=0&&im4[r-1][c+1]>lt)
		{
			im2[r-1][c+1]=0;
			Connect(r-1, c+1, lt, im2, im4);
		}
		
		if(im2[r+1][c]!=0&&im4[r+1][c]>lt)
		{
			im2[r+1][c]=0;
			Connect(r+1, c, lt, im2, im4);
		}
		
		if(im2[r+1][c-1]!=0&&im4[r+1][c-1]>lt)
		{
			im2[r+1][c-1]=0;
			Connect(r+1, c-1, lt, im2, im4);
		}
		
		if(im2[r+1][c+1]!=0&&im4[r+1][c+1]>lt)
		{
			im2[r+1][c+1]=0;
			Connect(r+1, c+1, lt, im2, im4);
		}
		if(im2[r][c+1]!=0&&im4[r][c+1]>lt)
		{
			im2[r][c+1]=0;
			Connect(r, c+1, lt, im2, im4);
		}
		if(im2[r][c-1]!=0&&im4[r][c-1]>lt)
		{
			im2[r][c-1]=0;
			Connect(r, c-1, lt, im2, im4);
		}
	}
}

void PgmImage::WritePGM(char *filename)
{
	int r;
	
	FILE *fptr;
	
	char rowSize[10];
	char colSize[10];
	char Color[10];
	
	sprintf(rowSize,"%d",row);
	sprintf(colSize,"%d",col);
	sprintf(Color,"%d",255);
	
	fptr = fopen(filename,"wb");		//create target file
	
	fputs("P5",fptr);					//format for pgm
	fputc('\n',fptr);
	
	fputs("#this is a pgm file in binaray format",fptr);	//comments
	fputc('\n',fptr); 
	
	fputs(colSize,fptr);				//write number of rows and cols and max value of grey level
	fputc(' ',fptr);
	
	fputs(rowSize,fptr);
	fputc('\n',fptr);
	
	fputs(Color,fptr);		
	fputc('\n',fptr);
	
	for (r=0; r<row; r++)				//write the file
		fwrite(img[r],sizeof(uchar),col,fptr);
	//	for (c=0; c<col; c++)
	//		fputc((uchar)(img[r][c]) ,fptr);
	fclose(fptr);						//close the file pointer
}




void PgmImage::DrawPoint(int x, int y)
{
	if(y>1 && y<row-2 && x>1 && x<col-2)
	{
		img[y+2][x+2]=255;
		img[y+2][x+1]=255;
		img[y+2][x]=255;
		img[y+2][x-1]=255;
		img[y+2][x-2]=255;

		img[y+1][x+2]=255;
		img[y][x+2]=255;
		img[y-1][x+2]=255;

		img[y-2][x+2]=255;
		img[y-2][x+1]=255;
		img[y-2][x]=255;
		img[y-2][x-1]=255;
		img[y-2][x-2]=255;

		img[y+1][x-2]=255;
		img[y][x-2]=255;
		img[y-1][x-2]=255;
	}
}

void PgmImage::DrawCross(int x, int y, uchar color)
{
	if(y>1 && y<row-2 && x>1 && x<col-2)
	{
		img[y][x+2]=color;
		img[y][x+1]=color;
		img[y][x]  =color;
		img[y][x-1]=color;
		img[y][x-2]=color;

		img[y+2][x]=color;
		img[y+1][x]=color;
		img[y-1][x]=color;
		img[y-2][x]=color;
	}
}

void PgmImage::WritePixel(int x, int y, uchar color)
{
	if(y>=0 && y<row && x>=0 && x<col)
	{
		img[y][x]=color;
	}
}

