//Written by Jiangjian Xiao

#include <afxwin.h>         // MFC core and standard components
#include <afxext.h>       // MFC extensions
#include <afxdtctl.h>		 // MFC support for Internet Explorer 4 Common Controls
#include <afxcmn.h>		 // MFC support for Windows Common Controls
#include "img.h"
#include "corner.h"
#include "com_fun.h"

PpmImage::PpmImage(int _row,int _col) 
{
	int r,c;
	row=_row;
	col=_col;
	img = new uchar*[row];		//assign memory to source and target arrays
	colorLevel=255;
	
	int length=3*col;
	for (r=0; r<row; r++)		//initialize image to 0
	{
		img[r]=new uchar[length];
		for (c=0; c<length; c++)
		{
			img[r][c] = 0;
		}
	}
}

PpmImage::PpmImage(int _row,int _col, Rgb &color) 
{
	int r,c;
	row=_row;
	col=_col;
	img = new uchar*[row];		//assign memory to source and target arrays
	colorLevel=255;
	
	for (r=0; r<row; r++)		//initialize image to 0
	{
		img[r]=new uchar[3*col];
		for (c=0; c<col; c++)
		{
			int idx=3*c;
			img[r][idx] = color.r;
			img[r][idx+1] = color.g;
			img[r][idx+2] = color.b;
		}
	}
}

PpmImage::PpmImage(PpmImage *_img)
{
	int r,c;
	row=_img->row;
	col=_img->col;
	img = new uchar*[row];		//assign memory to source and target arrays
	
	colorLevel=_img->colorLevel;
	int length=3*col;
	for (r=0; r<row; r++)		//initialize image to 0
	{
		img[r]=new uchar[length];
		for (c=0; c<length; c++)
		{
			img[r][c] = _img->img[r][c];
		}
	}
}

PpmImage::PpmImage()
{
	row=-1;
	col=-1;
	img=NULL;
}

PpmImage::PpmImage(char *filename)
{
	ReadPPM(filename);
}

PpmImage::~PpmImage()
{
	int i;

	if(row!=-1)
	{
		for(i=0; i<row; i++)
		{
			delete img[i];
		}
		delete img;
	}
}


void PpmImage::ReadPPM(char *filename)
{
	FILE *fp;
	int r;
	char buf[256];

	if ((fp = fopen(filename,"rb")) == NULL)
	{
		printf("File does not exist!\n");	//bad source file name
		exit(0);
	}
	else
	{
		if(fgets(buf,256,fp)==NULL || (buf[0]!='P' && buf[1]!='6'))
		//check if the source file is in correct format
		{	
			printf("Not a PPM file!\n");
			fclose(fp);
			return;
		}

		while(1)
		{
			fgets(buf,256,fp);	//comments are stored in char array comments
			if(buf[0]!='#')
			{
				sscanf(buf,"%d %d",&col,&row);
				break;
			}
		}
		
		fgets(buf,256,fp);		//read max grey level
		sscanf(buf,"%d",&colorLevel);
		
		img = new uchar*[row];		//assign memory to source and target arrays
		int length=3*col;
		for (r=0; r<row; r++)
		{
			img[r]=new uchar[length];
			fread(img[r],sizeof(uchar),length,fp);
		}
		
		fclose(fp);					//close the file pointer
	}
}

void PpmImage::WritePPM(char *filename)
{
	int r;
	
	FILE *fp;
	
	char rowSize[10];
	char colSize[10];
	char color[10];
	
	sprintf(rowSize,"%d",row);
	sprintf(colSize,"%d",col);
	sprintf(color,"%d",colorLevel);
	
	fp = fopen(filename,"wb");		//create target file
	
	fputs("P6",fp);					//format for pgm
	fputc('\n',fp);
	
	fputs("#this is a pgm file in binaray format",fp);	//comments
	fputc('\n',fp); 
	
	fputs(colSize,fp);				//write number of rows and cols and max value of grey level
	fputc(' ',fp);
	
	fputs(rowSize,fp);
	fputc('\n',fp);
	
	fputs(color,fp);		
	fputc('\n',fp);
	
	int length=3*col;
	for (r=0; r<row; r++)				//write the file
		fwrite(img[r],sizeof(uchar),length,fp);
	fclose(fp);						//close the file pointer
}


void PpmImage::Copy(PpmImage* _img)
{
	int r,c;

	int length=3*col;
	for (r=0; r<row; r++)		//initialize image to 0
	{
		for (c=0; c<length; c++)
		{
			img[r][c] = _img->img[r][c];
		}
	}
}

void PpmImage::Clear()
{
	int r,c;

	int length=3*col;
	for (r=0; r<row; r++)		//initialize image to 0
	{
		for (c=0; c<length; c++)
		{
			img[r][c] = 0;
		}
	}
}

void PpmImage::Clear(Rgb &color)
{
	int r,c;

	for (r=0; r<row; r++)		//initialize image to 0
	{
		for (c=0; c<col; c++)
		{
			int idx=3*c;
			img[r][idx] = color.r;
			img[r][idx+1] = color.g;
			img[r][idx+2] = color.b;
		}
	}
}

void PpmImage::WritePixel(int x, int y, Rgb &color)
{
	int idx=3*x;
	img[y][idx] = color.r;
	img[y][idx+1] = color.g;
	img[y][idx+2] = color.b;
}

void PpmImage::DrawPoint(int x, int y, Rgb &color)
{
	if(y>1 && y<row-2 && x>1 && x<col-2)
	{
		WritePixel(x+2,y+2,color);
		WritePixel(x+1,y+2,color);
		WritePixel(x,y+2,color);
		WritePixel(x-1,y+2,color);
		WritePixel(x-2,y+2,color);

		WritePixel(x+2,y+1,color);
		WritePixel(x+2,y,color);
		WritePixel(x+2,y-1,color);

		WritePixel(x+2,y-2,color);
		WritePixel(x+1,y-2,color);
		WritePixel(x,y-2,color);
		WritePixel(x-1,y-2,color);
		WritePixel(x-2,y-2,color);

		WritePixel(x-2,y+1,color);
		WritePixel(x-2,y,color);
		WritePixel(x-2,y-1,color);
	}
}


void PpmImage::Ppm2Pgm(PgmImage *pgm)
{
	int r,c;
	for(r=0; r<row;r++)
	{
		for(c=0;c<col;c++)
		{
			int n=3*c;
			//img->img[r][c]=round(0.333*cb+0.334*cg+0.333*cr);
			//pgm->img[r][c]=round(0.114*img[r][n]+0.587*img[r][n+1]+0.299*img[r][n+2]);
			pgm->img[r][c]=round(0.3811*img[r][n]+0.5783*img[r][n+1]+0.0402*img[r][n+2]);
		}
	}
}


